/*
 * Crasher -- Cause the kernel to crash/oops on command.
 *
 * Copyright (C) 2009 Canonical Ltd <lieb@canonical.com>
 *
 * This module provides a set of sysfs attributes that when written
 * to, cause the kernel to fault in the named way.
 *
 * This module is for TESTING and/or fault injection use *ONLY*!
 * There are NO security checks in this module.  Writing to any of
 * the attributes exported will DO BAD THINGS.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Changelog:
 *
 * 1 September 2006 - Jim Lieb <lieb@infoblox.com>
 *   V0.01 release
 */

#include <linux/types.h>
#include <linux/init.h>
#include <linux/err.h>
#include <linux/module.h>
#include <linux/string.h>
#include <linux/sysfs.h>
#include <linux/kobject.h>
#include <linux/kernel.h>
#include <linux/sched.h>

#define CRASHER_VERSION "0.01"
#define CRASHER_DATE "2009-Jan-30"

MODULE_AUTHOR("Jim Lieb <lieb@canonical.com>");
MODULE_DESCRIPTION("Crasher kernel OOPS/Panic crash trigger interface");
MODULE_LICENSE("GPL");
MODULE_VERSION(CRASHER_VERSION);

/* expose the triggers via /sys/kernel/crash  */


#define CRASH_ATTR_RW(_name) \
static struct kobj_attribute _name##_attr = \
	__ATTR(_name, 0644, _name##_show, _name##_store)

/* arm the crasher */

static int crash_arm;

static ssize_t arm_show(struct kobject *kobj,
			struct kobj_attribute *attr, char *buf)
{
	return sprintf(buf, "%u\n", crash_arm);
}

static ssize_t arm_store(struct kobject *kobj,
			 struct kobj_attribute *attr,
			 const char *buf, size_t count)
{
	crash_arm = simple_strtoul(buf, NULL, 0);
	return count;
}

CRASH_ATTR_RW(arm);

/* cause an OOPS by segfaulting */

static int crash_oops;

static ssize_t oops_show(struct kobject *kobj,
			struct kobj_attribute *attr, char *buf)
{
	return sprintf(buf, "%u\n", crash_oops);
}

static ssize_t oops_store(struct kobject *kobj,
			 struct kobj_attribute *attr,
			 const char *buf, size_t count)
{
	int *page0 = 0x0L;
	int tmp = 0xDEADBEEF;

	crash_oops = simple_strtoul(buf, NULL, 0);
	if(crash_arm && crash_oops) {
		printk(KERN_WARNING "Crasher caused a a segfault! pid=%u\n",
		       current->pid);
		*page0 = tmp; /* this is as far as we go... */
	}
	return count;
}

CRASH_ATTR_RW(oops);

/* cause a BUG dump */

static int crash_bug;

static ssize_t bug_show(struct kobject *kobj,
			struct kobj_attribute *attr, char *buf)
{
	return sprintf(buf, "%u\n", crash_bug);
}

static ssize_t bug_store(struct kobject *kobj,
			 struct kobj_attribute *attr,
			 const char *buf, size_t count)
{
	crash_bug = simple_strtoul(buf, NULL, 0);
	if(crash_arm && crash_bug) {
		printk(KERN_WARNING "Crasher caused a BUG dump! pid=%u\n",
		       current->pid);
		BUG(); /* this is as far as we go... */
	}
	return count;
}

CRASH_ATTR_RW(bug);

/* cause a WARN dump */

static int crash_warn;

static ssize_t warn_show(struct kobject *kobj,
			struct kobj_attribute *attr, char *buf)
{
	return sprintf(buf, "%u\n", crash_warn);
}

static ssize_t warn_store(struct kobject *kobj,
			 struct kobj_attribute *attr,
			 const char *buf, size_t count)
{
	crash_warn = simple_strtoul(buf, NULL, 0);
	if(crash_arm && crash_warn) {
		WARN(1, "Crasher caused a WARN dump! pid=%u\n",
		       current->pid);
	}
	return count;
}

CRASH_ATTR_RW(warn);

/* cause a PANIC dump */

static int crash_panic;

static ssize_t panic_show(struct kobject *kobj,
			struct kobj_attribute *attr, char *buf)
{
	return sprintf(buf, "%u\n", crash_panic);
}

static ssize_t panic_store(struct kobject *kobj,
			 struct kobj_attribute *attr,
			 const char *buf, size_t count)
{
	crash_panic = simple_strtoul(buf, NULL, 0);
	if(crash_arm && crash_panic) {
		panic("Crasher caused a panic! pid=%u\n",
		      current->pid);
	}
	return count;
}

CRASH_ATTR_RW(panic);

static struct attribute *crash_attrs[] = {
	&arm_attr.attr,
	&oops_attr.attr,
	&bug_attr.attr,
	&warn_attr.attr,
	&panic_attr.attr,
	NULL
};

static struct attribute_group crash_attr_group = {
	.attrs = crash_attrs,
	.name = "crash"
};

static int __init
crasher_init(void)
{
	int error = 0;

	printk(KERN_INFO "Kernel Crash Trigger v%s %s\n", CRASHER_VERSION,
	       CRASHER_DATE);

	error = sysfs_create_group(kernel_kobj, &crash_attr_group);
	return error;
}

static void __exit
crasher_exit(void)
{
	sysfs_remove_group(kernel_kobj, &crash_attr_group);
}

module_init(crasher_init);
module_exit(crasher_exit);
